#!/usr/bin/env python

import zipfile
import StringIO
import os.path

import OpenDocMill.Reader

class ODTFileTemplate(object):
    def __init__(self, inZipFilename):
        self.inZipFilename = inZipFilename
        self.contentTemplate = None
        self.stylesTemplate = None
        self.imageList = []
    
    def setContentTemplate(self, contentTemplate): self.contentTemplate = contentTemplate
    def setStylesTemplate(self, stylesTemplate): self.stylesTemplate = stylesTemplate
    def appendImage(self, filename):
        self.imageList.append(filename)

    def write(self, outZipFilename, data):
        inZipFile = zipfile.ZipFile(self.inZipFilename, "r")
        outZipFile = zipfile.ZipFile(outZipFilename, "w")
        for fileInfo in inZipFile.filelist:
            if fileInfo.filename == "content.xml" and self.contentTemplate is not None:
                s = StringIO.StringIO()
                self.contentTemplate.write(s, data, self.appendImage)
                _make_utf8_evilly(s)
                outZipFile.writestr(fileInfo, s.getvalue())
            elif fileInfo.filename == "styles.xml" and self.stylesTemplate is not None:
                s = StringIO.StringIO()
                self.stylesTemplate.write(s, data, self.appendImage)
                _make_utf8_evilly(s)
                outZipFile.writestr(fileInfo, s.getvalue())
            else:
                outZipFile.writestr(fileInfo, inZipFile.read(fileInfo.filename))
        for filename in self.imageList:
            basename = os.path.basename(filename)
            outZipFile.write(filename, arcname="Pictures/%s" % basename)
        outZipFile.close()

class XMLFileTemplate(object):
    def __init__(self, identifier, appendImage):
        self.identifier = identifier
        self.beforeText = []
        self.sections = {}
        self.afterText = []
        self.appendImage = appendImage

    def addBeforeText(self, text): self.beforeText.append(text)

    def addSection(self, name, section):
        if self.sections.has_key(name): raise ValueError, "Duplicate section name"
        self.sections[name] = section

    def addAfterText(self, text): self.afterText.append(text)

    def __repr__(self):
        return '\n'.join(
            ["TXTB: %r" % x for x in self.beforeText]
            + ["SECT %s: %r" % s for s in self.sections.items()]
            + ["TXTA: %r" % x for x in self.afterText])

    def write(self, stream, data, appendImage):
        stream.writelines(self.beforeText)
        for sectionData in data:
            if self.ignore(sectionData): continue
            sectionName = sectionData['name']
            try:
                section = self.sections[sectionName]
            except KeyError:
                raise ValueError, "No section %r in template %r" % (sectionName, self.identifier)
            section.write(stream,
                sectionData.get('fields', {}),
                sectionData.get('tables', {}),
                sectionData.get('images', {}),
                appendImage,
            )
        stream.writelines(self.afterText)

    def ignore(self, sectionData):
        raise NotImplementedError

class ContentTemplate(XMLFileTemplate):
    def ignore(self, sectionData):
        return sectionData.get('name').startswith('#')

class StylesTemplate(XMLFileTemplate):
    def ignore(self, sectionData):
        return sectionData.get('name') not in ['#header', '#footer']


class Section(object):
    def __init__(self, identifier):
        self.identifier = identifier
        self.elements = []

    def addText(self, text): self.elements.append(("TEXT", text))
    def addVariable(self, varName): self.elements.append(("VARIABLE", varName))
    def addTable(self, tableName, table): self.elements.append(("TABLE", (tableName, table)))
    def addImage(self, imageName, defaultArcFilename): self.elements.append(("IMAGE", (imageName, defaultArcFilename)))

    def __repr__(self):
        return '\n'.join([repr(x) for x in self.elements])

    def write(self, stream, fieldData, tableData, imageData, appendImage):
        for eType, e in self.elements:
            if eType == "TEXT":
                stream.write(e)
            elif eType == "VARIABLE":
                try:
                    rawString = unicode(fieldData[e])
                except KeyError:
                    raise ValueError, "No value for field %r in section %r" % (e, self.identifier)
                stream.write(xmlEscape(rawString))
            elif eType == "IMAGE":
                imageName, defaultArcfilename = e
                filename = imageData.get(imageName)
                if filename is None:
                    rawArcFilename = defaultArcFilename
                else:
                    rawArcFilename = u"Pictures/%s" % os.path.basename(filename)
                    appendImage(filename)
                stream.write(xmlEscapeAttr(rawArcFilename))
            elif eType == "TABLE":
                tableName, table = e
                try:
                    data = tableData[tableName]
                except KeyError:
                    raise ValueError, "No data for table %r in section %r" % (tableName, self.identifier)
                table.write(stream, data)
            else:
                raise ValueError, "Unknown type %r in template, section %r " % (eType, self.identifier)


class Table(object):
    def __init__(self, identifier):
        self.identifier = identifier
        self.beforeText = []
        self.row = None
        self.afterText = []

    def addBeforeText(self, text): self.beforeText.append(text)
    def setRow(self, row): self.row = row
    def addAfterText(self, text): self.afterText.append(text)

    def __repr__(self):
        return '\n'.join(
            ["    TBLB: %r" % x for x in self.beforeText]
            + ["    ROW : %r" % self.row]
            + ["    TBLA: %r" % x for x in self.afterText])

    def write(self, stream, data):
        stream.writelines(self.beforeText)
        for i in xrange(len(data)):
            rowFields = data[i]
            self.row.write(stream, rowFields, rowNo=i)
        stream.writelines(self.afterText)


class Row(object):
    def __init__(self, tableIdentifier):
        self.tableIdentifier = tableIdentifier
        self.elements = []

    def addText(self, text): self.elements.append(("TEXT", text))
    def addVariable(self, varName): self.elements.append(("VARIABLE", varName))

    def __repr__(self):
        return '\n'.join(["    %r" % (x,) for x in self.elements])

    def write(self, stream, fields, rowNo):
        for eType, e in self.elements:
            if eType == "TEXT":
                stream.write(e)
            elif eType == "VARIABLE":
                try:
                    rawString = fields[e]
                except KeyError:
                    raise ValueError, "No value for field %r in table %r[row=%d]" % (e, self.tableIdentifier, rowNo)
                stream.write(xmlEscape(rawString))
            else:
                raise ValueError, "Unknown type %r in template, table %r" % (eType, self.tableIdentifier)


def xmlEscape(s):
    return s.replace('&', '&amp;').replace('<', '&lt;')
def xmlEscapeAttr(s):
    return s.replace('&', '&amp;').replace('<', '&lt;').replace('"', '&quot;')
def _make_utf8_evilly(stringIO):
    b = stringIO.buflist
    for i in range(len(b)):
        if isinstance(b[i], unicode): b[i] = b[i].encode("UTF-8")

